@@ -1,10 +1,11 @@
+
 from rpython.jit.codewriter.effectinfo import EffectInfo
 from rpython.jit.metainterp.optimizeopt.optimizer import Optimization
-from rpython.jit.metainterp.resoperation import rop
+from rpython.jit.metainterp.resoperation import rop, OpHelpers
 
 
 def is_raw_free(op, opnum):
-    if opnum != rop.CALL:
+    if not OpHelpers.is_real_call(opnum):
         return False
     einfo = op.getdescr().get_extra_info()
     return einfo.oopspecindex == EffectInfo.OS_RAW_FREE
@@ -18,15 +19,14 @@ class OptEarlyForce(Optimization):
             opnum != rop.SETARRAYITEM_GC and
             opnum != rop.SETARRAYITEM_RAW and
             opnum != rop.QUASIIMMUT_FIELD and
-            opnum != rop.SAME_AS and
-            opnum != rop.MARK_OPAQUE_PTR and
+            opnum != rop.SAME_AS_I and
+            opnum != rop.SAME_AS_R and
+            opnum != rop.SAME_AS_F and
             not is_raw_free(op, opnum)):
 
             for arg in op.getarglist():
-                if arg in self.optimizer.values:
-                    value = self.getvalue(arg)
-                    value.force_box(self)
-        self.emit_operation(op)
+                self.optimizer.force_box(arg, self)
+        return self.emit(op)
 
     def setup(self):
         self.optimizer.optearlyforce = self
